<?php
// vim: set ai ts=4 sw=4 ft=php:
//	License for all code of this FreePBX module can be found in the license file inside the module directory
//	Copyright 2014 Schmooze Com Inc.
//
namespace FreePBX\modules;
use Symfony\Component\Process\Exception\ProcessFailedException;
use Symfony\Component\Process\Process;
class Ucpnode extends \FreePBX_Helpers implements \BMO {
	private $nodever = "0.10.40";
	private $npmver = "1.4.28";
	private $icuver = "4.2.1";
	private $foreverroot = "/tmp";
	private $nodeloc = "/tmp";
	public function __construct($freepbx = null) {
		$this->db = $freepbx->Database;
		$this->freepbx = $freepbx;
		$this->foreverroot = $this->freepbx->Config->get('ASTVARLIBDIR') . "/ucp";
		$this->nodeloc = __DIR__."/node";
		if(!file_exists($this->nodeloc."/logs")) {
			mkdir($this->nodeloc."/logs");
		}
	}

	public function install() {
		global $amp_conf;
		if(file_exists($amp_conf['AMPBIN']."/freepbx_engine_hook_ucpnode") && is_writable($amp_conf['AMPBIN']."/freepbx_engine_hook_ucpnode")) {
			unlink($amp_conf['AMPBIN']."/freepbx_engine_hook_ucpnode");
		}

		$output = exec("node --version"); //v0.10.29
		$output = str_replace("v","",trim($output));
		if(empty($output)) {
			out(_("Node is not installed"));
			return false;
		}
		if(version_compare($output,$this->nodever,"<")) {
			out(sprintf(_("Node version is: %s requirement is %s. Run 'yum upgrade nodejs' from the CLI as root"),$output,$this->nodever));
			return false;
		}


		$output = exec("npm --version"); //v0.10.29
		$output = trim($output);
		if(empty($output)) {
			out(_("Node Package Manager is not installed"));
			return false;
		}
		if(version_compare($output,$this->npmver,"<")) {
			out(sprintf(_("NPM version is: %s requirement is %s. Run 'yum upgrade nodejs' from the CLI as root"),$output,$this->npmver));
			return false;
		}

		$output = exec("icu-config --version"); //v4.2.1
		$output = trim($output);
		if(empty($output)) {
			out(_("icu is not installed. You need to run: yum install icu libicu-devel"));
			return false;
		}
		if(version_compare($output,$this->icuver,"<")) {
			out(sprintf(_("ICU version is: %s requirement is %s"),$output,$this->icuver));
			return false;
		}

		$cwd = getcwd();
		$webuser = \FreePBX::Freepbx_conf()->get('AMPASTERISKWEBUSER');
		chdir($this->nodeloc);
		putenv("PATH=/bin:/usr/bin:/sbin");
		putenv("USER=".$webuser);
		$web = posix_getpwnam($webuser);
		$home = trim($web['dir']);
		if (!is_dir($home)) {
			// Well, that's handy. It doesn't exist. Let's use ASTSPOOLDIR instead, because
			// that should exist and be writable.
			$home = \FreePBX::Freepbx_conf()->get('ASTSPOOLDIR');
			if (!is_dir($home)) {
				// OK, I give up.
				throw new \Exception(sprintf(_("Asterisk home dir (%s) doesn't exist, and, ASTSPOOLDIR doesn't exist. Aborting"),$home));
			}
		}

		putenv("HOME=".$home);
		putenv("FOREVER_ROOT=".$this->foreverroot);
		$astlogdir = $this->freepbx->Config->get("ASTLOGDIR");
		$varlibdir = $this->freepbx->Config->get("ASTVARLIBDIR");
		putenv("ASTLOGDIR=".$astlogdir);
		putenv("SHELL=/bin/bash");
		outn(_("Installing/Updating Required Libraries. This may take a while..."));
		file_put_contents($this->nodeloc."/logs/install.log","");
		$handle = popen("npm update 2>&1", "r");
		$log = fopen($this->nodeloc."/logs/install.log", "a");
		while (($buffer = fgets($handle, 4096)) !== false) {
			//out(trim($buffer));
			fwrite($log,$buffer);
			outn(".");
		}
		fclose($log);
		out("");
		out(_("Finished updating libraries!"));
		if(!file_exists($this->nodeloc."/node_modules/forever/bin/forever")) {
			out("");
			out(sprintf(_("There was an error installing. Please review the install log. (%s)"),$this->nodeloc."/logs/install.log"));
			return false;
		}
		if(!is_executable($this->nodeloc."/node_modules/forever/bin/forever")) {
			chmod($this->nodeloc."/node_modules/forever/bin/forever",0755);
		}

		$set = array();
		$set['module'] = 'ucpnode';
		$set['category'] = 'UCP NodeJS Server';

		// NODEJSENABLED
		$set['value'] = true;
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'Enable the NodeJS Server';
		$set['description'] = 'Whether to enable the backend server for UCP which allows instantaneous updates to the interface';
		$set['emptyok'] = 0;
		$set['level'] = 1;
		$set['readonly'] = 0;
		$set['type'] = CONF_TYPE_BOOL;
		$this->freepbx->Config->define_conf_setting('NODEJSENABLED',$set);

		// NODEJSTLSENABLED
		$set['value'] = false;
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'Enable TLS for the NodeJS Server';
		$set['description'] = 'Whether to enable TLS for the backend server for UCP which allows instantaneous updates to the interface';
		$set['emptyok'] = 0;
		$set['level'] = 1;
		$set['readonly'] = 0;
		$set['type'] = CONF_TYPE_BOOL;
		$this->freepbx->Config->define_conf_setting('NODEJSTLSENABLED',$set);

		// NODEJSBINDADDRESS
		$set['value'] = '::';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS Bind Address';
		$set['description'] = 'Address to bind to. Default is "::" (Listen for all IPv4 and IPv6 Connections)';
		$set['emptyok'] = 0;
		$set['type'] = CONF_TYPE_TEXT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSBINDADDRESS',$set);

		// NODEJSBINDPORT
		$set['value'] = '8001';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS Bind Port';
		$set['description'] = 'Port to bind to. Default is 8001';
		$set['emptyok'] = 0;
		$set['options'] = array(10,65536);
		$set['type'] = CONF_TYPE_INT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSBINDPORT',$set);

		// NODEJSHTTPSBINDADDRESS
		$set['value'] = '::';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS HTTPS Bind Address';
		$set['description'] = 'Address to bind to. Default is "::" (Listen for all IPv4 and IPv6 Connections)';
		$set['emptyok'] = 0;
		$set['type'] = CONF_TYPE_TEXT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSHTTPSBINDADDRESS',$set);

		// NODEJSHTTPSBINDPORT
		$set['value'] = '8003';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS HTTPS Bind Port';
		$set['description'] = 'Port to bind to. Default is 8003';
		$set['emptyok'] = 0;
		$set['options'] = array(10,65536);
		$set['type'] = CONF_TYPE_INT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSHTTPSBINDPORT',$set);

		// NODEJSTLSCERTFILE
		$set['value'] = '';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS HTTPS TLS Certificate Location';
		$set['description'] = 'Sets the path to the HTTPS server certificate. This is required if "Enable TLS for the NodeJS Server" is set to yes.';
		$set['emptyok'] = 1;
		$set['type'] = CONF_TYPE_TEXT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSTLSCERTFILE',$set);

		// NODEJSTLSPRIVATEKEY
		$set['value'] = '';
		$set['defaultval'] =& $set['value'];
		$set['options'] = '';
		$set['name'] = 'NodeJS HTTPS TLS Private Key Location';
		$set['description'] = 'Sets the path to the HTTPS private key. This is required if "Enable TLS for the NodeJS Server" is set to yes.';
		$set['emptyok'] = 1;
		$set['type'] = CONF_TYPE_TEXT;
		$set['level'] = 2;
		$set['readonly'] = 0;
		$this->freepbx->Config->define_conf_setting('NODEJSTLSPRIVATEKEY',$set);

		$this->freepbx->Config->commit_conf_settings();

		$cert = $this->freepbx->Certman->getDefaultCertDetails();
		if(!empty($cert)) {
			$this->setDefaultCert($cert);
		}

		if(!file_exists($varlibdir.'/ucp')) {
			mkdir($varlibdir.'/ucp');
		}
		//need forever to be executable. it's sooo cutable
		if(!is_executable($this->nodeloc."/node_modules/forever/bin/forever")) {
			chmod($this->nodeloc."/node_modules/forever/bin/forever",0755);
		}

		//If we are root then start it as asterisk, otherwise we arent root so start it as the web user (all we can do really)
		outn(_("Stopping old running processes..."));
		$this->stopFreepbx();
		out(_("Done"));

		outn(_("Starting new UCP Node Process..."));
		$started = $this->startFreepbx();
		if(!$started) {
			out(_("Failed!"));
			return false;
		} else {
			out(_("Started!"));
			return true;
		}
	}
	public function uninstall() {
		outn(_("Stopping old running processes..."));
		$this->stopFreepbx();
		out(_("Done"));
		exec("rm -Rf ".$this->nodeloc."/node_modules");
	}
	public function backup(){

	}
	public function restore($backup){

	}

	public function setDefaultCert($details) {
		$certF = $details['integration']['files']['crt'];
		$keyF = $details['integration']['files']['key'];
		$this->freepbx->Config->update("NODEJSTLSENABLED",true);
		$this->freepbx->Config->update("NODEJSTLSCERTFILE",$certF);
		$this->freepbx->Config->update("NODEJSTLSPRIVATEKEY",$keyF);
	}

	public function getHomeDir() {
		$webuser = \FreePBX::Freepbx_conf()->get('AMPASTERISKWEBUSER');
		$web = posix_getpwnam($webuser);
		$home = trim($web['dir']);
		if (!is_dir($home)) {
			// Well, that's handy. It doesn't exist. Let's use ASTSPOOLDIR instead, because
			// that should exist and be writable.
			$home = \FreePBX::Freepbx_conf()->get('ASTSPOOLDIR');
			if (!is_dir($home)) {
				// OK, I give up.
				throw new \Exception(sprintf(_("Asterisk home dir (%s) doesn't exist, and, ASTSPOOLDIR doesn't exist. Aborting"),$home));
			}
		}
		return $home;
	}

	public function dashboardService() {
		putenv("FOREVER_ROOT=".$this->foreverroot);
		putenv("HOME=".$this->getHomeDir());
		$services = array(
			array(
				'title' => _('UCP Daemon'),
				'type' => 'unknown',
				'tooltip' => _("Unknown"),
				'order' => 999,
				'command' => $this->nodeloc."/node_modules/forever/bin/forever list --plain"
			)
		);
		foreach($services as &$service) {
			$output = '';
			exec($service['command']." 2>&1", $output, $ret);

			if ($ret !== 0) {
				if(!file_exists($this->nodeloc."/node_modules/forever")) {
					$service = array_merge($service, $this->genAlertGlyphicon('critical', _("UCP Node is damaged. Please reinstall.")));
					continue;
				}
				$service = array_merge($service, $this->genAlertGlyphicon('critical', _("Unable to launch UCP Node monitor")));
				continue;
			}

			foreach($output as $line) {
				if(preg_match('/ucp/i',$line) && (preg_match('/ucp\.log(.*)/',$line,$matches) || preg_match('/ucp_forever\.log(.*)/',$line,$matches))) {
					$status = trim($matches[1]);
					if(preg_match("/STOPPED/",$status)) {
						$service = array_merge($service, $this->genAlertGlyphicon('critical', _("UCP Node is not running")));
					} else {
						$uptime = preg_replace('/\x1B\[([0-9]{1,3}((;[0-9]{1,3})*)?)?[m|K]/','',trim($matches[1]));
						$service = array_merge($service, $this->genAlertGlyphicon('ok', sprintf(_("Running (Uptime: %s)"),$uptime)));
					}
					continue(2);
				}
			}

			$service = array_merge($service, $this->genAlertGlyphicon('critical', _("UCP Node is not running")));
		}

		return $services;
	}

	private function genAlertGlyphicon($res, $tt = null) {
		$glyphs = array(
			"ok" => "glyphicon-ok text-success",
			"warning" => "glyphicon-warning-sign text-warning",
			"error" => "glyphicon-remove text-danger",
			"unknown" => "glyphicon-question-sign text-info",
			"info" => "glyphicon-info-sign text-info",
			"critical" => "glyphicon-fire text-danger"
		);
		// Are we being asked for an alert we actually know about?
		if (!isset($glyphs[$res])) {
			return array('type' => 'unknown', "tooltip" => sprintf(_("Don't know what %s is").$res), "glyph-class" => $glyphs['unknown']);
		}

		if ($tt === null) {
			// No Tooltip
			return array('type' => $res, "tooltip" => null, "glyph-class" => $glyphs[$res]);
		} else {
			// Generate a tooltip
			$html = '';
			if (is_array($tt)) {
				foreach ($tt as $line) {
					$html .= htmlentities($line, ENT_QUOTES, "UTF-8")."\n";
				}
			} else {
				$html .= htmlentities($tt, ENT_QUOTES, "UTF-8");
			}

			return array('type' => $res, "tooltip" => $html, "glyph-class" => $glyphs[$res]);
		}
		return '';
	}

	/**
	 * Start FreePBX for fwconsole hook
	 * @param object $output The output object.
	 */
	public function startFreepbx($output=null) {
		if(!$this->freepbx->Config->get("NODEJSENABLED")) {
			return;
		}
		$webroot = $this->freepbx->Config->get("AMPWEBROOT");
		$varlibdir = $this->freepbx->Config->get("ASTVARLIBDIR");
		$astlogdir = $this->freepbx->Config->get("ASTLOGDIR");

		if(!file_exists($webroot."/admin/modules/ucpnode/node/node_modules/forever")) {
			if(is_object($output)) {
				$output->writeln("<error>"._("UCP Node is damaged. Please reinstall.")."</error>");
			}
			return false;
		}

		if(!is_executable($webroot."/admin/modules/ucpnode/node/node_modules/forever")) {
			if(is_object($output)) {
				$output->writeln("<error>"._("Unable to launch UCP Node monitor")."</error>");
			}
			return false;
		}

		$cmds = array(
			'cd '.$webroot.'/admin/modules/ucpnode/node',
			'mkdir -p '.$varlibdir.'/ucp',
			'mkdir -p logs',
			'export FOREVER_ROOT='.$varlibdir.'/ucp',
			'export ASTLOGDIR='.$astlogdir,
			'export HOME='.$this->getHomeDir()
		);
		$cmds[] = "node_modules/forever/bin/forever list  --uid 'ucp'";
		$command = implode(" && ", $cmds);
		if (posix_getuid() == 0) {
			$process = new Process('runuser -l asterisk -c "'.$command.'"');
		} else {
			$process = new Process($command);
		}
		$process->run();
		if ($process->isSuccessful()) {
			$o = $process->getOutput();
			$o = explode("\n",$o);
			$time = '';
			foreach($o as $line) {
				if(preg_match('/ucp/i',$line) && (preg_match('/ucp\.log(.*)/',$line,$matches) || preg_match('/ucp_forever\.log(.*)/',$line,$matches))) {
					$status = trim($matches[1]);
					if(preg_match("/STOPPED/",$status)) {
						break;
					} else {
						$time = trim($status);
					}
				}
			}
			if(!empty($time)) {
				if(is_object($output)) {
					$output->writeln(sprintf(_("UCP Server has already been running for %s"),$time));
				}
				return true;
			}
		}

		array_pop($cmds);
		$cmds[] = 'npm start';
		$command = implode(" && ", $cmds);
		if (posix_getuid() == 0) {
			$command = 'runuser -l asterisk -c "'.$command.'"';
		}
		$process = new Process($command);
		$process->run();

		// executes after the command finishes
		if(is_object($output)) {
			$output->writeln(_("Starting UCP Server"));
		}
		if ($process->isSuccessful()) {
			if(is_object($output)) {
				$output->writeln(_("Started UCP Server"));
			}
		} else {
			if(is_object($output)) {
				$output->writeln("<error>".sprintf(_("UCP Server Failed to run: '%s'")."</error>",$command));
				return false;
			}
		}
		return true;
	}

	/**
	 * Stop FreePBX for fwconsole hook
	 * @param object $output The output object.
	 */
	public function stopFreepbx($output=null) {
		$webroot = $this->freepbx->Config->get("AMPWEBROOT");
		$varlibdir = $this->freepbx->Config->get("ASTVARLIBDIR");
		$astlogdir = $this->freepbx->Config->get("ASTLOGDIR");
		if(!file_exists($webroot."/admin/modules/ucpnode/node/node_modules/forever")) {
			if(is_object($output)) {
				$output->writeln("<error>"._("UCP Node is damaged. Please reinstall.")."</error>");
			}
			return false;
		}

		if(!is_executable($webroot."/admin/modules/ucpnode/node/node_modules/forever")) {
			if(is_object($output)) {
				$output->writeln("<error>"._("Unable to launch UCP Node monitor")."</error>");
			}
			return false;
		}

		$cmds = array(
			'cd '.$webroot.'/admin/modules/ucpnode/node',
			'mkdir -p '.$varlibdir.'/ucp',
			'mkdir -p logs',
			'export FOREVER_ROOT='.$varlibdir.'/ucp',
			'export ASTLOGDIR='.$astlogdir,
			'export HOME='.$this->getHomeDir()
		);

		$cmds[] = 'npm stop';
		$command = implode(" && ", $cmds);
		if (posix_getuid() == 0) {
			$process = new Process('runuser -l asterisk -c "'.$command.'"');
		} else {
			$process = new Process($command);
		}
		$process->run();

		// executes after the command finishes
		if(is_object($output)) {
			$output->writeln(_("Stopping UCP Server"));
		}
		if ($process->isSuccessful()) {
			if(is_object($output)) {
				$output->writeln(_("Stopped UCP Server"));
			}
		} else {
			if(is_object($output)) {
				$output->writeln("<error>".sprintf(_("UCP Server Failed: %s")."</error>",$process->getErrorOutput()));
				return false;
			}
		}
		return true;
	}

	/**
	 * FreePBX chown hooks
	 */
	public function chownFreepbx() {
		$files = array();

		$files[] = array('type' => 'execdir',
			'path' => __DIR__.'/node/node_modules/forever/bin',
			'perms' => 0755);

		$files[] = array('type' => 'rdir',
			'path' => $this->freepbx->Config->get("ASTVARLIBDIR").'/ucp',
			'perms' => 0755);

		return $files;
	}
}
